/*
 * RFC2367 PF_KEYv2 Key management API message parser
 * Copyright (C) 1999  Richard Guy Briggs.
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * RCSID $Id: pfkey_v2_parser.c,v 1.10 1999/09/21 15:24:45 rgb Exp $
 */

/*
 *		Template from klips/net/ipsec/ipsec/netlink.c.
 */

char pfkey_msg_parse_c_version[] = "$Id: pfkey_v2_parser.c,v 1.10 1999/09/21 15:24:45 rgb Exp $";

#include <linux/config.h>
#include <linux/version.h>

#include <linux/kernel.h> /* printk() */
#include <linux/malloc.h> /* kmalloc() */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/interrupt.h> /* mark_bh */

#include <linux/netdevice.h>   /* struct device, and other headers */
#include <linux/etherdevice.h> /* eth_type_trans */
#include <linux/ip.h>          /* struct iphdr */
#include <linux/skbuff.h>
#include "../../../lib/freeswan.h"
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
#include <asm/uaccess.h>
#include <linux/in6.h>
#define ip_chk_addr inet_addr_type
#define IS_MYADDR RTN_LOCAL
#endif
#include <asm/checksum.h>
#include <net/ip.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
#include <net/netlink.h>
#else
#include <linux/netlink.h>
#endif

#include <linux/random.h>	/* get_random_bytes() */

#include "radij.h"
#include "ipsec_encap.h"
#include "ipsec_radij.h"
#include "ipsec_netlink.h"
#include "ipsec_xform.h"
#include "ipsec_ah.h"
#include "ipsec_esp.h"

#include "pfkey.h"
#include "pfkeyv2.h"

#define SENDERR(_x) do { error = -(_x); goto errlab; } while (0)

#ifndef min
#define min(a,b)	(((a)<(b))?(a):(b))
#endif

extern int des_set_key(caddr_t, caddr_t);

struct sklist_t {
	struct socket *sk;
	struct sklist_t* next;
} pfkey_sklist_head, *pfkey_sklist, *pfkey_sklist_prev;

unsigned int exts_permitted_in[SADB_MAX + 1] = {
0
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_EXT_ADDRESS_PROXY
| 1<<SADB_EXT_SPIRANGE
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_LIFETIME_CURRENT
| 1<<SADB_EXT_LIFETIME_HARD
| 1<<SADB_EXT_LIFETIME_SOFT
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_EXT_ADDRESS_PROXY
| 1<<SADB_EXT_KEY_AUTH
| 1<<SADB_EXT_KEY_ENCRYPT
| 1<<SADB_EXT_IDENTITY_SRC
| 1<<SADB_EXT_IDENTITY_DST
| 1<<SADB_EXT_SENSITIVITY
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_LIFETIME_HARD
| 1<<SADB_EXT_LIFETIME_SOFT
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_EXT_ADDRESS_PROXY
| 1<<SADB_EXT_KEY_AUTH
| 1<<SADB_EXT_KEY_ENCRYPT
| 1<<SADB_EXT_IDENTITY_SRC
| 1<<SADB_EXT_IDENTITY_DST
| 1<<SADB_EXT_SENSITIVITY
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_EXT_ADDRESS_PROXY
| 1<<SADB_EXT_IDENTITY_SRC
| 1<<SADB_EXT_IDENTITY_DST
| 1<<SADB_EXT_SENSITIVITY
| 1<<SADB_EXT_PROPOSAL
,
1<<SADB_EXT_RESERVED
,
0
,
1<<SADB_EXT_RESERVED
,
1<<SADB_EXT_RESERVED
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_LIFETIME_CURRENT
| 1<<SADB_EXT_LIFETIME_HARD
| 1<<SADB_EXT_LIFETIME_SOFT
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_EXT_ADDRESS_PROXY
| 1<<SADB_EXT_KEY_AUTH
| 1<<SADB_EXT_KEY_ENCRYPT
| 1<<SADB_EXT_IDENTITY_SRC
| 1<<SADB_EXT_IDENTITY_DST
| 1<<SADB_EXT_SENSITIVITY
| 1<<SADB_EXT_PROPOSAL
| 1<<SADB_EXT_SUPPORTED_AUTH
| 1<<SADB_EXT_SUPPORTED_ENCRYPT
| 1<<SADB_EXT_SPIRANGE
| 1<<SADB_X_EXT_KMPRIVATE
| 1<<SADB_X_EXT_SATYPE2
| 1<<SADB_X_EXT_SA2
| 1<<SADB_X_EXT_ADDRESS_DST2
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_LIFETIME_CURRENT
| 1<<SADB_EXT_LIFETIME_HARD
| 1<<SADB_EXT_LIFETIME_SOFT
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_EXT_ADDRESS_PROXY
| 1<<SADB_EXT_KEY_AUTH
| 1<<SADB_EXT_KEY_ENCRYPT
| 1<<SADB_EXT_IDENTITY_SRC
| 1<<SADB_EXT_IDENTITY_DST
| 1<<SADB_EXT_SENSITIVITY
| 1<<SADB_EXT_PROPOSAL
| 1<<SADB_EXT_SUPPORTED_AUTH
| 1<<SADB_EXT_SUPPORTED_ENCRYPT
| 1<<SADB_EXT_SPIRANGE
| 1<<SADB_X_EXT_KMPRIVATE
| 1<<SADB_X_EXT_SATYPE2
| 1<<SADB_X_EXT_SA2
| 1<<SADB_X_EXT_ADDRESS_DST2
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_X_EXT_SATYPE2
| 1<<SADB_X_EXT_SA2
| 1<<SADB_X_EXT_ADDRESS_DST2
};

unsigned int exts_required_in[SADB_MAX + 1] = {
0
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_EXT_ADDRESS_PROXY
| 1<<SADB_EXT_SPIRANGE
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_EXT_KEY_AUTH
| 1<<SADB_EXT_KEY_ENCRYPT
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
/*| 1<<SADB_EXT_KEY_AUTH*/
/*| 1<<SADB_EXT_KEY_ENCRYPT*/
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_EXT_PROPOSAL
,
1<<SADB_EXT_RESERVED
,
0
,
1<<SADB_EXT_RESERVED
,
1<<SADB_EXT_RESERVED
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_LIFETIME_CURRENT
| 1<<SADB_EXT_LIFETIME_HARD
| 1<<SADB_EXT_LIFETIME_SOFT
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_EXT_ADDRESS_PROXY
| 1<<SADB_EXT_KEY_AUTH
| 1<<SADB_EXT_KEY_ENCRYPT
| 1<<SADB_EXT_IDENTITY_SRC
| 1<<SADB_EXT_IDENTITY_DST
| 1<<SADB_EXT_SENSITIVITY
| 1<<SADB_EXT_PROPOSAL
| 1<<SADB_EXT_SUPPORTED_AUTH
| 1<<SADB_EXT_SUPPORTED_ENCRYPT
| 1<<SADB_EXT_SPIRANGE
| 1<<SADB_X_EXT_KMPRIVATE
| 1<<SADB_X_EXT_SATYPE2
| 1<<SADB_X_EXT_SA2
| 1<<SADB_X_EXT_ADDRESS_DST2
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_LIFETIME_CURRENT
| 1<<SADB_EXT_LIFETIME_HARD
| 1<<SADB_EXT_LIFETIME_SOFT
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_EXT_ADDRESS_PROXY
| 1<<SADB_EXT_KEY_AUTH
| 1<<SADB_EXT_KEY_ENCRYPT
| 1<<SADB_EXT_IDENTITY_SRC
| 1<<SADB_EXT_IDENTITY_DST
| 1<<SADB_EXT_SENSITIVITY
| 1<<SADB_EXT_PROPOSAL
| 1<<SADB_EXT_SUPPORTED_AUTH
| 1<<SADB_EXT_SUPPORTED_ENCRYPT
| 1<<SADB_EXT_SPIRANGE
| 1<<SADB_X_EXT_KMPRIVATE
| 1<<SADB_X_EXT_SATYPE2
| 1<<SADB_X_EXT_SA2
| 1<<SADB_X_EXT_ADDRESS_DST2
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_X_EXT_SATYPE2
| 1<<SADB_X_EXT_SA2
| 1<<SADB_X_EXT_ADDRESS_DST2
};

unsigned int exts_permitted_out[SADB_MAX + 1] = {
0
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_LIFETIME_CURRENT
| 1<<SADB_EXT_LIFETIME_HARD
| 1<<SADB_EXT_LIFETIME_SOFT
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_EXT_ADDRESS_PROXY
| 1<<SADB_EXT_IDENTITY_SRC
| 1<<SADB_EXT_IDENTITY_DST
| 1<<SADB_EXT_SENSITIVITY
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_LIFETIME_HARD
| 1<<SADB_EXT_LIFETIME_SOFT
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_EXT_IDENTITY_SRC
| 1<<SADB_EXT_IDENTITY_DST
| 1<<SADB_EXT_SENSITIVITY
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_LIFETIME_CURRENT
| 1<<SADB_EXT_LIFETIME_HARD
| 1<<SADB_EXT_LIFETIME_SOFT
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_EXT_ADDRESS_PROXY
| 1<<SADB_EXT_KEY_AUTH
| 1<<SADB_EXT_KEY_ENCRYPT
| 1<<SADB_EXT_IDENTITY_SRC
| 1<<SADB_EXT_IDENTITY_DST
| 1<<SADB_EXT_SENSITIVITY
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_EXT_ADDRESS_PROXY
| 1<<SADB_EXT_IDENTITY_SRC
| 1<<SADB_EXT_IDENTITY_DST
| 1<<SADB_EXT_SENSITIVITY
| 1<<SADB_EXT_PROPOSAL
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SUPPORTED_AUTH
| 1<<SADB_EXT_SUPPORTED_ENCRYPT
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_LIFETIME_CURRENT
| 1<<SADB_EXT_LIFETIME_HARD
| 1<<SADB_EXT_LIFETIME_SOFT
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
,
1<<SADB_EXT_RESERVED
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_LIFETIME_CURRENT
| 1<<SADB_EXT_LIFETIME_HARD
| 1<<SADB_EXT_LIFETIME_SOFT
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_EXT_ADDRESS_PROXY
| 1<<SADB_EXT_KEY_AUTH
| 1<<SADB_EXT_KEY_ENCRYPT
| 1<<SADB_EXT_IDENTITY_SRC
| 1<<SADB_EXT_IDENTITY_DST
| 1<<SADB_EXT_SENSITIVITY
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_LIFETIME_CURRENT
| 1<<SADB_EXT_LIFETIME_HARD
| 1<<SADB_EXT_LIFETIME_SOFT
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_EXT_ADDRESS_PROXY
| 1<<SADB_EXT_KEY_AUTH
| 1<<SADB_EXT_KEY_ENCRYPT
| 1<<SADB_EXT_IDENTITY_SRC
| 1<<SADB_EXT_IDENTITY_DST
| 1<<SADB_EXT_SENSITIVITY
| 1<<SADB_EXT_PROPOSAL
| 1<<SADB_EXT_SUPPORTED_AUTH
| 1<<SADB_EXT_SUPPORTED_ENCRYPT
| 1<<SADB_EXT_SPIRANGE
| 1<<SADB_X_EXT_KMPRIVATE
| 1<<SADB_X_EXT_SATYPE2
| 1<<SADB_X_EXT_SA2
| 1<<SADB_X_EXT_ADDRESS_DST2
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_LIFETIME_CURRENT
| 1<<SADB_EXT_LIFETIME_HARD
| 1<<SADB_EXT_LIFETIME_SOFT
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_EXT_ADDRESS_PROXY
| 1<<SADB_EXT_KEY_AUTH
| 1<<SADB_EXT_KEY_ENCRYPT
| 1<<SADB_EXT_IDENTITY_SRC
| 1<<SADB_EXT_IDENTITY_DST
| 1<<SADB_EXT_SENSITIVITY
| 1<<SADB_EXT_PROPOSAL
| 1<<SADB_EXT_SUPPORTED_AUTH
| 1<<SADB_EXT_SUPPORTED_ENCRYPT
| 1<<SADB_EXT_SPIRANGE
| 1<<SADB_X_EXT_KMPRIVATE
| 1<<SADB_X_EXT_SATYPE2
| 1<<SADB_X_EXT_SA2
| 1<<SADB_X_EXT_ADDRESS_DST2
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_X_EXT_SATYPE2
| 1<<SADB_X_EXT_SA2
| 1<<SADB_X_EXT_ADDRESS_DST2
};

unsigned int exts_required_out[SADB_MAX + 1] = {
0
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_EXT_KEY_AUTH
| 1<<SADB_EXT_KEY_ENCRYPT
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_EXT_PROPOSAL
,
1<<SADB_EXT_RESERVED
/* | 1<<SADB_EXT_SUPPORTED_AUTH
   | 1<<SADB_EXT_SUPPORTED_ENCRYPT */
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_LIFETIME_CURRENT
/* | 1<<SADB_EXT_LIFETIME_HARD
   | 1<<SADB_EXT_LIFETIME_SOFT */
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
,
1<<SADB_EXT_RESERVED
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_EXT_KEY_AUTH
| 1<<SADB_EXT_KEY_ENCRYPT
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_LIFETIME_CURRENT
| 1<<SADB_EXT_LIFETIME_HARD
| 1<<SADB_EXT_LIFETIME_SOFT
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_EXT_ADDRESS_PROXY
| 1<<SADB_EXT_KEY_AUTH
| 1<<SADB_EXT_KEY_ENCRYPT
| 1<<SADB_EXT_IDENTITY_SRC
| 1<<SADB_EXT_IDENTITY_DST
| 1<<SADB_EXT_SENSITIVITY
| 1<<SADB_EXT_PROPOSAL
| 1<<SADB_EXT_SUPPORTED_AUTH
| 1<<SADB_EXT_SUPPORTED_ENCRYPT
| 1<<SADB_EXT_SPIRANGE
| 1<<SADB_X_EXT_KMPRIVATE
| 1<<SADB_X_EXT_SATYPE2
| 1<<SADB_X_EXT_SA2
| 1<<SADB_X_EXT_ADDRESS_DST2
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_LIFETIME_CURRENT
| 1<<SADB_EXT_LIFETIME_HARD
| 1<<SADB_EXT_LIFETIME_SOFT
| 1<<SADB_EXT_ADDRESS_SRC
| 1<<SADB_EXT_ADDRESS_DST
| 1<<SADB_EXT_ADDRESS_PROXY
| 1<<SADB_EXT_KEY_AUTH
| 1<<SADB_EXT_KEY_ENCRYPT
| 1<<SADB_EXT_IDENTITY_SRC
| 1<<SADB_EXT_IDENTITY_DST
| 1<<SADB_EXT_SENSITIVITY
| 1<<SADB_EXT_PROPOSAL
| 1<<SADB_EXT_SUPPORTED_AUTH
| 1<<SADB_EXT_SUPPORTED_ENCRYPT
| 1<<SADB_EXT_SPIRANGE
| 1<<SADB_X_EXT_KMPRIVATE
| 1<<SADB_X_EXT_SATYPE2
| 1<<SADB_X_EXT_SA2
| 1<<SADB_X_EXT_ADDRESS_DST2
,
1<<SADB_EXT_RESERVED
| 1<<SADB_EXT_SA
| 1<<SADB_EXT_ADDRESS_DST
};

__u8 sadb_satype2proto[] = {
0,
0,
IPPROTO_AH,
IPPROTO_ESP,
0,
0,
0,
0,
0,
IPPROTO_IPIP
};

static int
pfkey_sa_parse(struct sock *sk, struct sadb_msg *pfkey_msg,
	       struct sadb_ext *pfkey_ext, struct tdb *tdb1, struct tdb *tdb2)
{
	int error = 0;
	struct tdb *tdbp;
	struct sadb_sa *pfkey_sa = (struct sadb_sa *)pfkey_ext;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_sa_parse: \n");
	/* sanity checks... */
	if(!pfkey_sa) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sa_parse: NULL pointer passed in.\n");
		SENDERR(EINVAL);
	}

	if(pfkey_sa->sadb_sa_len != sizeof(struct sadb_sa) / IPSEC_PFKEYv2_ALIGN) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sa_parse: length wrong pfkey_sa->sadb_sa_len=%d sizeof(struct sadb_sa)=%d.\n",
			    pfkey_sa->sadb_sa_len,
			    sizeof(struct sadb_sa));
		SENDERR(EINVAL);
	}

	if(pfkey_sa->sadb_sa_encrypt > SADB_EALG_MAX) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sa_parse: pfkey_sa->sadb_sa_encrypt=%d > SADB_EALG_MAX=%d.\n",
			    pfkey_sa->sadb_sa_encrypt,
			    SADB_EALG_MAX);
		SENDERR(EINVAL);
	}

	if(pfkey_sa->sadb_sa_auth > SADB_AALG_MAX) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sa_parse: pfkey_sa->sadb_sa_auth=%d > SADB_AALG_MAX=%d.\n",
			    pfkey_sa->sadb_sa_auth,
			    SADB_AALG_MAX);
		SENDERR(EINVAL);
	}

#if 0
	if(!pfkey_sa->sadb_sa_encrypt && !pfkey_sa->sadb_sa_auth) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sa_parse: enc:both algos zero: pfkey_sa->sadb_sa_encrypt=%d, pfkey_sa->sadb_sa_auth=%d.\n",
			    pfkey_sa->sadb_sa_encrypt,
			    pfkey_sa->sadb_sa_auth);
		SENDERR(EINVAL);
	}
#endif

	if(pfkey_sa->sadb_sa_state > SADB_SASTATE_MAX) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sa_parse: state=%d exceeds MAX=%d.\n",
			    pfkey_sa->sadb_sa_state,
			    SADB_SASTATE_MAX);
		SENDERR(EINVAL);
	}

	if(pfkey_sa->sadb_sa_state == SADB_SASTATE_DEAD) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sa_parse: state=%d is DEAD=%d.\n",
			    pfkey_sa->sadb_sa_state,
			    SADB_SASTATE_DEAD);
		SENDERR(EINVAL);
	}

	if((pfkey_msg->sadb_msg_type == SADB_ADD) &&
	   (pfkey_sa->sadb_sa_state != SADB_SASTATE_MATURE)) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sa_parse: state=%d for add should be MATURE=%d.\n",
			    pfkey_sa->sadb_sa_state,
			    SADB_SASTATE_MATURE);
		SENDERR(EINVAL);
	}

	if((pfkey_msg->sadb_msg_type == SADB_UPDATE) &&
	   (pfkey_sa->sadb_sa_state != SADB_SASTATE_MATURE)) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sa_parse: state=%d for update should be MATURE=%d.\n",
			    pfkey_sa->sadb_sa_state,
			    SADB_SASTATE_MATURE);
		SENDERR(EINVAL);
	}

	switch(pfkey_sa->sadb_sa_exttype) {
	case SADB_EXT_SA:
		tdbp = tdb1;
		break;
	case SADB_X_EXT_SA2:
		tdbp = tdb2;
		break;
	default:
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sa_parse: unknown exttype=%d, expecting SADB_EXT_SA=%d or SADB_X_EXT_SA2=%d.\n",
			    pfkey_sa->sadb_sa_exttype,
			    SADB_EXT_SA,
			    SADB_X_EXT_SA2);
		SENDERR(EINVAL);
	}

	tdbp->tdb_said.spi = pfkey_sa->sadb_sa_spi;
	if(pfkey_sa->sadb_sa_replay > 64) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sa_parse: replay window size: %d"
			    " -- must be 0 <= size <= 64\n",
			    pfkey_sa->sadb_sa_replay);
		SENDERR(EINVAL);
	}
	tdbp->tdb_replaywin = pfkey_sa->sadb_sa_replay;
	tdbp->tdb_state = pfkey_sa->sadb_sa_state;
	tdbp->tdb_flags = pfkey_sa->sadb_sa_flags;
	tdbp->tdb_replaywin_lastseq = tdbp->tdb_replaywin_bitmap = 0;

	switch(tdbp->tdb_said.proto) {
		case IPPROTO_AH:
			if(pfkey_sa->sadb_sa_encrypt != SADB_EALG_NONE) {
				KLIPS_PRINT(debug_pfkey,
					    "klips_debug:pfkey_sa_parse: AH handed encalg=%d.\n",
					    pfkey_sa->sadb_sa_encrypt);
				SENDERR(EINVAL);
			}

			tdbp->tdb_authalg = pfkey_sa->sadb_sa_auth;
			tdbp->tdb_encalg = SADB_EALG_NONE;

			break;
		case IPPROTO_ESP:
			if((pfkey_sa->sadb_sa_encrypt == SADB_EALG_NULL) &&
			   (pfkey_sa->sadb_sa_auth == SADB_AALG_NONE) ) {
				KLIPS_PRINT(debug_pfkey,
					    "klips_debug:pfkey_sa_parse: ESP handed encNULL+authNONE.\n");

				SENDERR(EINVAL);
			}

			tdbp->tdb_authalg = pfkey_sa->sadb_sa_auth;
			tdbp->tdb_encalg = pfkey_sa->sadb_sa_encrypt;

			break;
		case IPPROTO_IPIP:
			tdbp->tdb_authalg = AH_NONE;
			tdbp->tdb_encalg = ESP_NONE;
			break;
#ifdef CONFIG_IPSEC_COMP
		case IPPROTO_COMP:
			break;
#endif /* CONFIG_IPSEC_COMP */
		default:
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_sa_parse: Unknown protocol=%d.\n",
				    tdbp->tdb_said.proto);
			
			SENDERR(EINVAL);
	}

errlab:
	return error;
}

static int
pfkey_lifetime_parse(struct sock *sk, struct sadb_msg *pfkey_msg,
	       struct sadb_ext *pfkey_ext, struct tdb *tdbp, struct tdb *tdb2)
{
	int error = 0;
	struct sadb_lifetime *pfkey_lifetime = (struct sadb_lifetime *)pfkey_ext;
	struct timeval time;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_lifetime_parse: \n");
	/* sanity checks... */
	if(!pfkey_lifetime) {
		SENDERR(EINVAL);
	}

	if(pfkey_lifetime->sadb_lifetime_len != sizeof(struct sadb_lifetime) /
	   IPSEC_PFKEYv2_ALIGN) {
		SENDERR(EINVAL);
	}

	do_gettimeofday(&time);
	switch(pfkey_lifetime->sadb_lifetime_exttype) {
#if 0
	case SADB_EXT_LIFETIME_CURRENT:
		tdbp->tdb_lifetime_allocations_c = pfkey_lifetime->sadb_lifetime_allocations;
		tdbp->tdb_lifetime_bytes_c = pfkey_lifetime->sadb_lifetime_bytes;
		tdbp->tdb_lifetime_addtime_c = pfkey_lifetime->sadb_lifetime_addtime;
		tdbp->tdb_lifetime_usetime_c = pfkey_lifetime->sadb_lifetime_usetime;
		break;
#endif
	case SADB_EXT_LIFETIME_HARD:
		tdbp->tdb_lifetime_allocations_h = pfkey_lifetime->sadb_lifetime_allocations;
		tdbp->tdb_lifetime_bytes_h = pfkey_lifetime->sadb_lifetime_bytes;
		if((tdbp->tdb_lifetime_addtime_h = pfkey_lifetime->sadb_lifetime_addtime)) {
			tdbp->tdb_lifetime_addtime_h += time.tv_sec;
		}
		tdbp->tdb_lifetime_usetime_h = pfkey_lifetime->sadb_lifetime_usetime;
		break;
	case SADB_EXT_LIFETIME_SOFT:
		tdbp->tdb_lifetime_allocations_s = pfkey_lifetime->sadb_lifetime_allocations;
		tdbp->tdb_lifetime_bytes_s = pfkey_lifetime->sadb_lifetime_bytes;
		if((tdbp->tdb_lifetime_addtime_s = pfkey_lifetime->sadb_lifetime_addtime)) {
			tdbp->tdb_lifetime_addtime_s += time.tv_sec;
		}
		tdbp->tdb_lifetime_usetime_s = pfkey_lifetime->sadb_lifetime_usetime;
		break;
	default:
		SENDERR(EINVAL);
	}

errlab:
	return error;
}

static int
pfkey_address_parse(struct sock *sk, struct sadb_msg *pfkey_msg,
		    struct sadb_ext *pfkey_ext, struct tdb *tdb1, struct tdb *tdb2)
{
	int error = 0;
	int saddr_len = 0;
	char ipaddr_txt[16];
	unsigned char **sap;
	struct sadb_address *pfkey_address = (struct sadb_address *)pfkey_ext;
	struct sockaddr* s = (struct sockaddr*)((void*)pfkey_address + sizeof(*pfkey_address));
	
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_address_parse: \n");
	/* sanity checks... */
	if(!pfkey_address) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_address_parse: NULL pointer passed in.\n");
		SENDERR(EINVAL);
	}
	
	if(pfkey_address->sadb_address_len <
	   (sizeof(struct sadb_address) + sizeof(struct sockaddr))/
	   IPSEC_PFKEYv2_ALIGN) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_address_parse: size wrong ext_len=%d, adr_ext_len=%d, saddr_len=%d.\n",
			    pfkey_address->sadb_address_len,
			    sizeof(struct sadb_address),
			    sizeof(struct sockaddr));
		SENDERR(EINVAL);
	}
	
	if(pfkey_address->sadb_address_reserved) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_address_parse: res=%d.\n",
			    pfkey_address->sadb_address_reserved);
		SENDERR(EINVAL);
	}
	
	switch(s->sa_family) {
	case AF_INET:
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_address_parse: found address family=%d, AF_INET.\n",
			    s->sa_family);
		saddr_len = sizeof(struct sockaddr_in);
		addrtoa(((struct sockaddr_in*)s)->sin_addr, 0, ipaddr_txt, sizeof(ipaddr_txt));
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_address_parse: found address=%s.\n",
			    ipaddr_txt);
		break;
#if 0
	case AF_INET6:
		saddr_len = sizeof(struct sockaddr_in6);
		break;
#endif
	default:
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_address_parse: s->sa_family=%d not supported.\n",
			    s->sa_family);
		SENDERR(EPFNOSUPPORT);
	}
	
	if(pfkey_address->sadb_address_len !=
	   (sizeof(struct sadb_address) + saddr_len) / IPSEC_PFKEYv2_ALIGN) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_address_parse: size wrong ext_len=%d, adr_ext_len=%d, saddr_len=%d.\n",
			    pfkey_address->sadb_address_len,
			    sizeof(struct sadb_address),
			    saddr_len);
		SENDERR(EINVAL);
	}
	
	if(pfkey_address->sadb_address_prefixlen != 0) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_address_parse: address prefixes not supported yet.\n");
		SENDERR(EAFNOSUPPORT); /* not supported yet */
	}
	
	switch(pfkey_address->sadb_address_exttype) {
	case SADB_EXT_ADDRESS_SRC:
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_address_parse: found src address.\n");
		sap = (unsigned char **)&(tdb1->tdb_addr_s);
		tdb1->tdb_addr_s_size = saddr_len;
		break;
	case SADB_EXT_ADDRESS_DST:
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_address_parse: found dst address.\n");
		sap = (unsigned char **)&(tdb1->tdb_addr_d);
		tdb1->tdb_addr_d_size = saddr_len;
		break;
	case SADB_EXT_ADDRESS_PROXY:
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_address_parse: found proxy address.\n");
		sap = (unsigned char **)&(tdb1->tdb_addr_p);
		tdb1->tdb_addr_p_size = saddr_len;
		break;
	case SADB_X_EXT_ADDRESS_DST2:
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_address_parse: found 2nd dst address.\n");
		sap = (unsigned char **)&(tdb2->tdb_addr_d);
		tdb2->tdb_addr_d_size = saddr_len;
		break;
	default:
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_address_parse: unrecognised ext_type=%d.\n",
			    pfkey_address->sadb_address_exttype);
		SENDERR(EINVAL);
	}
	
	if(!(*sap = kmalloc(saddr_len, GFP_KERNEL))) {
		SENDERR(ENOMEM);
	}
	memcpy(*sap, s, saddr_len);
	
	if(pfkey_address->sadb_address_exttype == SADB_EXT_ADDRESS_DST) {
		if(s->sa_family == AF_INET) {
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_address_parse: tdb1->tdb_said.dst.s_addr=%08x, \n"
				    "klips_debug:                     &tdb1->tdb_said.dst.s_addr=%p, \n"
				    "klips_debug:                     ((struct sockaddr_in*)(tdb1->tdb_addr_d))->sin_addr.s_addr=%08x, \n"
				    "klips_debug:                     &((struct sockaddr_in*)(tdb1->tdb_addr_d))->sin_addr.s_addr=%p.\n",
				    tdb1->tdb_said.dst.s_addr,
				    &tdb1->tdb_said.dst.s_addr,
				    ((struct sockaddr_in*)(tdb1->tdb_addr_d))->sin_addr.s_addr,
				    &((struct sockaddr_in*)(tdb1->tdb_addr_d))->sin_addr.s_addr
				);
			tdb1->tdb_said.dst.s_addr = ((struct sockaddr_in*)(tdb1->tdb_addr_d))->sin_addr.s_addr;
			addrtoa(((struct sockaddr_in*)(tdb1->tdb_addr_d))->sin_addr/*tdb1->tdb_said.dst*/, 0, ipaddr_txt, sizeof(ipaddr_txt));
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_address_parse: tdb_said.dst set to %s.\n",
				    ipaddr_txt);
		} else {
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_address_parse: uh, tdb_said.dst doesn't do address family=%d yet, said will be invalid.\n",
				    s->sa_family);
		}
	}

	/* XXX check if port!=0 */
	
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_address_parse: successful.\n");
 errlab:
	return error;
}

static int
pfkey_key_parse(struct sock *sk, struct sadb_msg *pfkey_msg,
	       struct sadb_ext *pfkey_ext, struct tdb *tdbp, struct tdb *tdb2)
{
	int error = 0;
	struct sadb_key *pfkey_key = (struct sadb_key *)pfkey_ext;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_key_parse: \n");
	/* sanity checks... */

	if(!pfkey_key) {
		SENDERR(EINVAL);
	}

	if(pfkey_key->sadb_key_len < sizeof(struct sadb_key) / IPSEC_PFKEYv2_ALIGN) {
		SENDERR(EINVAL);
	}

	if(!pfkey_key->sadb_key_bits) {
		SENDERR(EINVAL);
	}

	if(pfkey_key->sadb_key_len != (sizeof(struct sadb_key) / IPSEC_PFKEYv2_ALIGN) +
	   ((pfkey_key->sadb_key_bits + 63) / 64)) {
		SENDERR(EINVAL);
	}

	if(pfkey_key->sadb_key_reserved) {
		SENDERR(EINVAL);
	}

	switch(pfkey_key->sadb_key_exttype) {
	case SADB_EXT_KEY_AUTH:
		tdbp->tdb_key_bits_a = pfkey_key->sadb_key_bits;
		if(!(tdbp->tdb_key_a = kmalloc((pfkey_key->sadb_key_bits + 7 / 8), GFP_KERNEL))) {
			SENDERR(ENOMEM);
		}
		memcpy(tdbp->tdb_key_a,
		       (char*)pfkey_key + sizeof(struct sadb_key),
		       (pfkey_key->sadb_key_bits + 7 ) / 8);
		break;
	case SADB_EXT_KEY_ENCRYPT: /* Key(s) */
		tdbp->tdb_key_bits_e = pfkey_key->sadb_key_bits;
		if(!(tdbp->tdb_key_e = kmalloc((pfkey_key->sadb_key_bits + 7 / 8), GFP_KERNEL))) {
			SENDERR(ENOMEM);
		}
		memcpy(tdbp->tdb_key_e,
		       (char*)pfkey_key + sizeof(struct sadb_key),
		       (pfkey_key->sadb_key_bits + 7 ) / 8);
		break;
	default:
		SENDERR(EINVAL);
 	}

errlab:
	return error;
}

static int
pfkey_ident_parse(struct sock *sk, struct sadb_msg *pfkey_msg,
	       struct sadb_ext *pfkey_ext, struct tdb *tdbp, struct tdb *tdb2)
{
	int error = 0;
	struct sadb_ident *pfkey_ident = (struct sadb_ident *)pfkey_ext;

	/* sanity checks... */
	if(pfkey_ident->sadb_ident_len < sizeof(struct sadb_ident) / IPSEC_PFKEYv2_ALIGN) {
		SENDERR(EINVAL);
	}

	if(pfkey_ident->sadb_ident_type > SADB_IDENTTYPE_MAX) {
		SENDERR(EINVAL);
	}

	if(pfkey_ident->sadb_ident_reserved) {
		SENDERR(EINVAL);
	}

	/* string terminator/padding must be zero */
	if(pfkey_ident->sadb_ident_len > sizeof(struct sadb_ident) / IPSEC_PFKEYv2_ALIGN) {
		if(*((char*)pfkey_ident + pfkey_ident->sadb_ident_len * IPSEC_PFKEYv2_ALIGN - 1)) {
			SENDERR(EINVAL);
		}
	}

	switch(pfkey_ident->sadb_ident_exttype) {
	case SADB_EXT_IDENTITY_SRC:
		tdbp->tdb_ident_type_s = pfkey_ident->sadb_ident_type;
		tdbp->tdb_ident_id_s = pfkey_ident->sadb_ident_id;
		tdbp->tdb_ident_len_s = pfkey_ident->sadb_ident_len -
			sizeof(struct sadb_ident) / IPSEC_PFKEYv2_ALIGN;
		if(!(tdbp->tdb_ident_data_s = kmalloc(pfkey_ident->sadb_ident_len * IPSEC_PFKEYv2_ALIGN, GFP_KERNEL))) {
			SENDERR(ENOMEM);
		}
		memcpy(tdbp->tdb_ident_data_s,
		       (char*)pfkey_ident + sizeof(struct sadb_ident),
		       pfkey_ident->sadb_ident_len * IPSEC_PFKEYv2_ALIGN);
		break;
	case SADB_EXT_IDENTITY_DST: /* Identity(ies) */
		tdbp->tdb_ident_type_d = pfkey_ident->sadb_ident_type;
		tdbp->tdb_ident_id_d = pfkey_ident->sadb_ident_id;
		tdbp->tdb_ident_len_d = pfkey_ident->sadb_ident_len -
			sizeof(struct sadb_ident) / IPSEC_PFKEYv2_ALIGN;
		if(!(tdbp->tdb_ident_data_d = kmalloc(pfkey_ident->sadb_ident_len * IPSEC_PFKEYv2_ALIGN, GFP_KERNEL))) {
			SENDERR(ENOMEM);
		}
		memcpy(tdbp->tdb_ident_data_d,
		       (char*)pfkey_ident + sizeof(struct sadb_ident),
		       pfkey_ident->sadb_ident_len * IPSEC_PFKEYv2_ALIGN);
		break;
	default:
		SENDERR(EINVAL);
 	}
errlab:
	return error;
}

static int
pfkey_sens_parse(struct sock *sk, struct sadb_msg *pfkey_msg,
	       struct sadb_ext *pfkey_ext, struct tdb *tdbp, struct tdb *tdb2)
{
	int error = 0;
	struct sadb_sens *pfkey_sens = (struct sadb_sens *)pfkey_ext;

	/* sanity checks... */
	if(pfkey_sens->sadb_sens_len < sizeof(struct sadb_sens) / IPSEC_PFKEYv2_ALIGN) {
		SENDERR(EINVAL);
	}

	SENDERR(EINVAL); /* don't process these yet */

errlab:
	return error;
}

static int
pfkey_prop_parse(struct sock *sk, struct sadb_msg *pfkey_msg,
	       struct sadb_ext *pfkey_ext, struct tdb *tdbp, struct tdb *tdb2)
{
	int error = 0;
	int i;
	struct sadb_prop *pfkey_prop = (struct sadb_prop *)pfkey_ext;

	/* sanity checks... */
	if(pfkey_prop->sadb_prop_len < sizeof(struct sadb_prop) / IPSEC_PFKEYv2_ALIGN) {
		SENDERR(EINVAL);
	}

	for(i=0; i<3; i++) {
		if(pfkey_prop->sadb_prop_reserved[i]) {
			SENDERR(EINVAL);
		}
	}

	SENDERR(EINVAL); /* don't process these yet */

errlab:
	return error;
}

static int
pfkey_supported_parse(struct sock *sk, struct sadb_msg *pfkey_msg,
	       struct sadb_ext *pfkey_ext, struct tdb *tdbp, struct tdb *tdb2)
{
	int error = 0;
	struct sadb_supported *pfkey_supported = (struct sadb_supported *)pfkey_ext;

	/* sanity checks... */
	if(pfkey_supported->sadb_supported_len < sizeof(struct sadb_supported) / IPSEC_PFKEYv2_ALIGN) {
		SENDERR(EINVAL);
	}

	if(pfkey_supported->sadb_supported_reserved) {
		SENDERR(EINVAL);
	}

	SENDERR(EINVAL); /* don't process these yet */

errlab:
	return error;
}

static int
pfkey_spirange_parse(struct sock *sk, struct sadb_msg *pfkey_msg,
	       struct sadb_ext *pfkey_ext, struct tdb *tdbp, struct tdb *tdb2)
{
	int error = 0;
	struct sadb_spirange *pfkey_spirange = (struct sadb_spirange *)pfkey_ext;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_spirange_parse: \n");
	/* sanity checks... */
	if(pfkey_spirange->sadb_spirange_len != sizeof(struct sadb_spirange) / IPSEC_PFKEYv2_ALIGN) {
		SENDERR(EINVAL);
	}

	if(pfkey_spirange->sadb_spirange_reserved) {
		SENDERR(EINVAL);
	}

	if(pfkey_spirange->sadb_spirange_max < pfkey_spirange->sadb_spirange_min) {
		SENDERR(EINVAL);
	}

	SENDERR(EINVAL); /* don't process these yet */

errlab:
	return error;
}

static int
pfkey_x_kmprivate_parse(struct sock *sk, struct sadb_msg *pfkey_msg,
	       struct sadb_ext *pfkey_ext, struct tdb *tdbp, struct tdb *tdb2)
{
	int error = 0;
	struct sadb_x_kmprivate *pfkey_x_kmprivate = (struct sadb_x_kmprivate *)pfkey_ext;

	/* sanity checks... */
	if(pfkey_x_kmprivate->sadb_x_kmprivate_len < sizeof(struct sadb_x_kmprivate) / IPSEC_PFKEYv2_ALIGN) {
		SENDERR(EINVAL);
	}

	if(pfkey_x_kmprivate->sadb_x_kmprivate_reserved) {
		SENDERR(EINVAL);
	}

	SENDERR(EINVAL); /* don't process these yet */

errlab:
	return error;
}

static int
pfkey_x_satype_parse(struct sock *sk, struct sadb_msg *pfkey_msg,
	       struct sadb_ext *pfkey_ext, struct tdb *tdb1, struct tdb *tdb2)
{
	int error = 0;
	int i;
	struct sadb_x_satype *pfkey_x_satype = (struct sadb_x_satype *)pfkey_ext;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_x_satype_parse: \n");
	/* sanity checks... */
	if(pfkey_x_satype->sadb_x_satype_len < sizeof(struct sadb_x_satype) / IPSEC_PFKEYv2_ALIGN) {
		SENDERR(EINVAL);
	}

	if(pfkey_x_satype->sadb_x_satype_satype > SADB_SATYPE_MAX) {
		SENDERR(EINVAL);
	}

	if(!pfkey_x_satype->sadb_x_satype_satype) {
		SENDERR(EINVAL);
	}

	for(i = 0; i < 3; i++) {
		if(pfkey_x_satype->sadb_x_satype_reserved[i]) {
			SENDERR(EINVAL);
		}
	}

	if(!(tdb2->tdb_proto = sadb_satype2proto[pfkey_x_satype->sadb_x_satype_satype])) {
		SENDERR(EINVAL);
	}

errlab:
	return error;
}

int (*ext_parsers[SADB_EXT_MAX +1])(struct sock*, struct sadb_msg*, struct sadb_ext*, struct tdb*, struct tdb*)
 =
{
	NULL, /* pfkey_msg_parse, */
	pfkey_sa_parse,
	pfkey_lifetime_parse,
	pfkey_lifetime_parse,
	pfkey_lifetime_parse,
	pfkey_address_parse,
	pfkey_address_parse,
	pfkey_address_parse,
	pfkey_key_parse,
	pfkey_key_parse,
	pfkey_ident_parse,
	pfkey_ident_parse,
	pfkey_sens_parse,
	pfkey_prop_parse,
	pfkey_supported_parse,
	pfkey_supported_parse,
	pfkey_spirange_parse,
	pfkey_x_kmprivate_parse,
	pfkey_x_satype_parse,
	pfkey_sa_parse,
	pfkey_address_parse
};

int
pfkey_tdb_init(struct tdb *tdbp, void **extensions)
{
        int i;
        int error = 0;
        char sa[SATOA_BUF];

	satoa(tdbp->tdb_said, 0, sa, SATOA_BUF);

        KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_tdb_init: (pfkey defined) called for SA:%s\n", sa);

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_tdb_init: calling init routine of %s%s%s\n",
		    TDB_XFORM_NAME(tdbp));
	
	switch(tdbp->tdb_proto) {
		
#ifdef CONFIG_IPSEC_IPIP
	case IPPROTO_IPIP: {
		if((tdbp->tdb_addr_s = (struct sockaddr*)
		    kmalloc((tdbp->tdb_addr_s_size = sizeof(struct sockaddr_in)),
			    GFP_ATOMIC)) == NULL) {
			SENDERR(ENOMEM);
		}
		if((tdbp->tdb_addr_d = (struct sockaddr*)
		    kmalloc((tdbp->tdb_addr_d_size = sizeof(struct sockaddr_in)),
			    GFP_ATOMIC)) == NULL) {
			SENDERR(ENOMEM);
		}
		
		/* might want to use a different structure here, or set sin_family and sin_port */
		((struct sockaddr_in*)(tdbp->tdb_addr_s))->sin_addr =
			((struct sockaddr_in*)
			 (extensions[SADB_EXT_ADDRESS_SRC] +
			  sizeof(struct sadb_address))
				)->sin_addr;
		((struct sockaddr_in*)(tdbp->tdb_addr_d))->sin_addr =
			((struct sockaddr_in*)
			 (extensions[SADB_EXT_ADDRESS_DST] +
			  sizeof(struct sadb_address))
				)->sin_addr;
	}
	break;
#endif /* !CONFIG_IPSEC_IPIP */
#ifdef CONFIG_IPSEC_AH
	case IPPROTO_AH:
		switch(tdbp->tdb_authalg) {
#ifdef CONFIG_IPSEC_AUTH_HMAC_MD5
		case AH_MD5: {
			unsigned char *akp;
			
			if(tdbp->tdb_key_bits_a != (AHMD596_KLEN * 8)) {
				KLIPS_PRINT(debug_pfkey,
					    "klips_debug:pfkey_tdb_init: incorrect key size: %d bits"
					    "-- must be %d bits\n"/*octets (bytes)\n"*/,
					    tdbp->tdb_key_bits_a, AHMD596_KLEN * 8);
				SENDERR(EINVAL);
			}
			
#if 0
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_tdb_init: hmac md5-96 key is 0x%08lx %08lx %08lx %08lx\n",
				    ntohl(*(((__u32 *)ed->ame_key)+0)),
				    ntohl(*(((__u32 *)ed->ame_key)+1)),
				    ntohl(*(((__u32 *)ed->ame_key)+2)),
				    ntohl(*(((__u32 *)ed->ame_key)+3)));
#endif
			
			tdbp->tdb_auth_bits = AHMD596_ALEN * 8;
			
			/* save the pointer to the key material */
			akp = tdbp->tdb_key_a;
			
			if((tdbp->tdb_key_a = (caddr_t)
			    kmalloc((tdbp->tdb_key_a_size = sizeof(struct md5_ctx)),
				    GFP_ATOMIC)) == NULL) {
				SENDERR(ENOMEM);
			}
			for (i = 0; i < (tdbp->tdb_key_bits_a + 7)/ 8; i++) {
				akp[i] ^= 0x36;
			}
			MD5Init(&((struct md5_ctx*)(tdbp->tdb_key_a))->ictx);
			
			MD5Update(&((struct md5_ctx*)(tdbp->tdb_key_a))->ictx,
				  akp,
				  (tdbp->tdb_key_bits_a + 7)/ 8);
			for (; i < AHMD596_BLKLEN; i++) {
				MD5Update(&((struct md5_ctx*)(tdbp->tdb_key_a))->ictx,
					  "\0x36", 1);
			}
			
			MD5Init(&((struct md5_ctx*)(tdbp->tdb_key_a))->octx);
			for (i = 0; i < (tdbp->tdb_key_bits_a + 7)/ 8; i++) {
				akp[i] ^= (0x36 ^ 0x5c);
			}
			MD5Update(&((struct md5_ctx*)(tdbp->tdb_key_a))->octx,
				  akp,
				  (tdbp->tdb_key_bits_a + 7)/ 8);
			for (; i < AHMD596_BLKLEN; i++) {
				MD5Update(&((struct md5_ctx*)(tdbp->tdb_key_a))->ictx,
					  "\0x5c", 1);
			}
			
#if 0
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_tdb_init: MD5 ictx=0x%08x %08x %08x %08x"
				    " octx=0x%08x %08x %08x %08x\n",
				    *((__u32*)&((struct md5_ctx*)(tdbp->tdb_key_a))->ictx),
				    *(((__u32*)&((struct md5_ctx*)(tdbp->tdb_key_a))->ictx) + 1),
				    *(((__u32*)&((struct md5_ctx*)(tdbp->tdb_key_a))->ictx) + 2),
				    *(((__u32*)&((struct md5_ctx*)(tdbp->tdb_key_a))->ictx) + 3),
				    *((__u32*)&((struct md5_ctx*)(tdbp->tdb_key_a))->octx),
				    *(((__u32*)&((struct md5_ctx*)(tdbp->tdb_key_a))->octx) + 1),
				    *(((__u32*)&((struct md5_ctx*)(tdbp->tdb_key_a))->octx) + 2),
				    *(((__u32*)&((struct md5_ctx*)(tdbp->tdb_key_a))->octx) + 3));
#endif
			
				/* zero key buffer -- paranoid */
			memset(akp, 0, (tdbp->tdb_key_bits_a + 7)/ 8);
		}
		break;
#endif /* CONFIG_IPSEC_AUTH_HMAC_MD5 */
#ifdef CONFIG_IPSEC_AUTH_HMAC_SHA1
		case AH_SHA: {
			
			unsigned char *akp;
			
			if(tdbp->tdb_key_bits_a != (AHSHA196_KLEN * 8)) {
				KLIPS_PRINT(debug_pfkey,
					    "klips_debug:pfkey_tdb_init: incorrect key size: %d bits"
					    "-- must be %d bits\n"/*octets (bytes)\n"*/,
					    tdbp->tdb_key_bits_a, AHSHA196_KLEN * 8);
				SENDERR(EINVAL);
			}
			
#if 0
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_tdb_init: hmac sha1-96 key is 0x%08lx %08lx %08lx %08lx\n",
				    ntohl(*(((__u32 *)ed->ame_key)+0)),
				    ntohl(*(((__u32 *)ed->ame_key)+1)),
				    ntohl(*(((__u32 *)ed->ame_key)+2)),
				    ntohl(*(((__u32 *)ed->ame_key)+3)));
#endif
			
			tdbp->tdb_auth_bits = AHSHA196_ALEN * 8;
			
			/* save the pointer to the key material */
			akp = tdbp->tdb_key_a;
			
			if((tdbp->tdb_key_a = (caddr_t)
			    kmalloc((tdbp->tdb_key_a_size = sizeof(struct sha1_ctx)),
				    GFP_ATOMIC)) == NULL) {
				SENDERR(ENOMEM);
					}
			for (i = 0; i < (tdbp->tdb_key_bits_a + 7)/ 8; i++) {
				akp[i] ^= 0x36;
			}
			SHA1Init(&((struct sha1_ctx*)(tdbp->tdb_key_a))->ictx);
			
			SHA1Update(&((struct sha1_ctx*)(tdbp->tdb_key_a))->ictx,
				   akp,
				   (tdbp->tdb_key_bits_a + 7)/ 8);
			for (; i < AHSHA196_BLKLEN; i++) {
				SHA1Update(&((struct sha1_ctx*)(tdbp->tdb_key_a))->ictx,
					   "\0x36", 1);
					}
			
			SHA1Init(&((struct sha1_ctx*)(tdbp->tdb_key_a))->octx);
			for (i = 0; i < (tdbp->tdb_key_bits_a + 7)/ 8; i++) {
				akp[i] ^= (0x36 ^ 0x5c);
			}
			SHA1Update(&((struct sha1_ctx*)(tdbp->tdb_key_a))->octx,
				   akp,
				   (tdbp->tdb_key_bits_a + 7)/ 8);
			for (; i < AHSHA196_BLKLEN; i++) {
				SHA1Update(&((struct sha1_ctx*)(tdbp->tdb_key_a))->ictx,
					   "\0x5c", 1);
			}
			
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_tdb_init: SHA1 ictx=0x%08x %08x %08x %08x"
				    " octx=0x%08x %08x %08x %08x\n", 
				    *((__u32*)&((struct sha1_ctx*)(tdbp->tdb_key_a))->ictx),
				    *(((__u32*)&((struct sha1_ctx*)(tdbp->tdb_key_a))->ictx) + 1),
				    *(((__u32*)&((struct sha1_ctx*)(tdbp->tdb_key_a))->ictx) + 2),
				    *(((__u32*)&((struct sha1_ctx*)(tdbp->tdb_key_a))->ictx) + 3),
				    *((__u32*)&((struct sha1_ctx*)(tdbp->tdb_key_a))->octx),
						    *(((__u32*)&((struct sha1_ctx*)(tdbp->tdb_key_a))->octx) + 1),
				    *(((__u32*)&((struct sha1_ctx*)(tdbp->tdb_key_a))->octx) + 2),
				    *(((__u32*)&((struct sha1_ctx*)(tdbp->tdb_key_a))->octx) + 3));
			
				/* zero key buffer -- paranoid */
			memset(akp, 0, (tdbp->tdb_key_bits_a + 7)/ 8);
		}
		break;
#endif /* CONFIG_IPSEC_AUTH_HMAC_SHA1 */
		default:
			SENDERR(EINVAL);
		}
#endif /* CONFIG_IPSEC_AH */
#ifdef CONFIG_IPSEC_ESP
	case IPPROTO_ESP: {
		unsigned char *akp, *ekp;
		
		switch(tdbp->tdb_encalg) {
#ifdef CONFIG_IPSEC_ENC_DES
		case ESP_DES:
#endif /* CONFIG_IPSEC_ENC_DES */
#ifdef CONFIG_IPSEC_ENC_3DES
		case ESP_3DES:
#endif /* CONFIG_IPSEC_ENC_3DES */
#if defined(CONFIG_IPSEC_ENC_DES) || defined(CONFIG_IPSEC_ENC_3DES)
			if((tdbp->tdb_iv = (caddr_t)
			    kmalloc((tdbp->tdb_iv_size = EMT_ESPDES_IV_SZ), GFP_ATOMIC)) == NULL) {
				SENDERR(ENOMEM);
			}
			get_random_bytes((void *)tdbp->tdb_iv, EMT_ESPDES_IV_SZ);
			break;
#endif /* defined(CONFIG_IPSEC_ENC_DES) || defined(CONFIG_IPSEC_ENC_3DES) */
		case ESP_NONE:
#ifdef CONFIG_IPSEC_ENC_NULL
		case ESP_NULL:
#endif /* CONFIG_IPSEC_ENC_NULL */
			break;
		default:
					SENDERR(EINVAL);
		}
		
		switch(tdbp->tdb_encalg) {
#ifdef CONFIG_IPSEC_ENC_DES
		case ESP_DES:
			if(tdbp->tdb_key_bits_e != (EMT_ESPDES_KEY_SZ * 8)) {
				KLIPS_PRINT(debug_pfkey,
					    "klips_debug:pfkey_tdb_init: incorrect encryption"
					    "key size: %d bits -- must be %d bits\n"/*octets (bytes)\n"*/,
					    tdbp->tdb_key_bits_e, EMT_ESPDES_KEY_SZ * 8);
				SENDERR(EINVAL);
			}
			
			/* save encryption key pointer */
			ekp = tdbp->tdb_key_e;
			
			if((tdbp->tdb_key_e = (caddr_t)
			    kmalloc((tdbp->tdb_key_e_size = sizeof(struct des_eks)),
				    GFP_ATOMIC)) == NULL) {
				SENDERR(ENOMEM);
			}
#if 0
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_tdb_init: des key 1 is 0x%08lx%08lx\n",
				    ntohl(*((__u32 *)ed->eme_key)),
				    ntohl(*((__u32 *)ed->eme_key + 1)));
#endif
			error = des_set_key((caddr_t)ekp, (caddr_t)(tdbp->tdb_key_e));
			if (error == -1)
				printk("klips_debug:pfkey_tdb_init: parity error in des key\n");
			else if (error == -2)
				printk("klips_debug:pfkey_tdb_init: illegal weak des key\n");
			if (error) {
				memset(tdbp->tdb_key_e, 0, sizeof(struct des_eks));
				kfree_s(tdbp->tdb_key_e, sizeof(struct des_eks));
				memset(ekp, 0, (tdbp->tdb_key_bits_e + 7)/ 8);
				SENDERR(EINVAL);
			}
			memset(ekp, 0, (tdbp->tdb_key_bits_e + 7)/ 8);
			break;
#endif /* CONFIG_IPSEC_ENC_DES */
#ifdef CONFIG_IPSEC_ENC_3DES
		case ESP_3DES:
			if(tdbp->tdb_key_bits_e != (EMT_ESP3DES_KEY_SZ * 8)) {
				KLIPS_PRINT(debug_pfkey,
					    "klips_debug:pfkey_tdb_init: incorrect encryption"
					    "key size: %d bits -- must be %d bits\n"/*octets (bytes)\n"*/,
					    tdbp->tdb_key_bits_e, EMT_ESP3DES_KEY_SZ * 8);
				SENDERR(EINVAL);
			}
			
			/* save encryption key pointer */
			ekp = tdbp->tdb_key_e;
			
			if((tdbp->tdb_key_e = (caddr_t)
			    kmalloc((tdbp->tdb_key_e_size = 3 * sizeof(struct des_eks)),
				    GFP_ATOMIC)) == NULL) {
				SENDERR(ENOMEM);
			}
			
			for(i = 0; i < 3; i++) {
#if 0
				KLIPS_PRINT(debug_pfkey,
					    "klips_debug:pfkey_tdb_init: 3des key %d/3 is 0x%08lx%08lx\n",
					    i + 1,
					    ntohl(*((__u32 *)ed->eme_key + i * 2)),
					    ntohl(*((__u32 *)ed->eme_key + i * 2 + 1)));
#endif
				error = des_set_key((caddr_t)ekp + EMT_ESPDES_KEY_SZ * i,
						    (caddr_t)&((struct des_eks*)(tdbp->tdb_key_e))[i]);
				if (error == -1)
					printk("klips_debug:pfkey_tdb_init: parity error in des key %d/3\n", i + 1);
				else if (error == -2)
					printk("klips_debug:pfkey_tdb_init: illegal weak des key %d/3\n", i + 1);
				if (error) {
					memset(tdbp->tdb_key_e, 0, 3 * sizeof(struct des_eks));
					kfree_s(tdbp->tdb_key_e, 3 * sizeof(struct des_eks));
					memset(ekp, 0, (tdbp->tdb_key_bits_e + 7)/ 8);
					SENDERR(EINVAL);
				}
			}

			/* paranoid */
			memset(ekp, 0, (tdbp->tdb_key_bits_e + 7)/ 8);
			
			break;
#endif /* CONFIG_IPSEC_ENC_3DES */
#ifdef CONFIG_IPSEC_ENC_NULL
		case ESP_NULL:
#endif /* CONFIG_IPSEC_ENC_NULL */
		case ESP_NONE:
			break;
		default:
			SENDERR(EINVAL);
		}
		
		switch(tdbp->tdb_authalg) {
#ifdef CONFIG_IPSEC_AUTH_HMAC_MD5
		case AH_MD5:
			if(tdbp->tdb_key_bits_a != (AHMD596_KLEN * 8)) {
				KLIPS_PRINT(debug_pfkey,
					    "klips_debug:pfkey_tdb_init: incorrect authorisation"
					    " key size: %d bits -- must be %d bits\n"/*octets (bytes)\n"*/,
					    tdbp->tdb_key_bits_a, AHMD596_KLEN * 8);
				SENDERR(EINVAL);
			}
			
#if 0
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_tdb_init: hmac md5-96 key is 0x%08lx %08lx %08lx %08lx\n",
				    ntohl(*(((__u32 *)ed->ame_key)+0)),
				    ntohl(*(((__u32 *)ed->ame_key)+1)),
				    ntohl(*(((__u32 *)ed->ame_key)+2)),
				    ntohl(*(((__u32 *)ed->ame_key)+3)));
#endif
			tdbp->tdb_auth_bits = AHMD596_ALEN * 8;
			
			/* save the pointer to the key material */
			akp = tdbp->tdb_key_a;
			
			if((tdbp->tdb_key_a = (caddr_t)
			    kmalloc((tdbp->tdb_key_a_size = sizeof(struct md5_ctx)),
				    GFP_ATOMIC)) == NULL) {
				SENDERR(ENOMEM);
			}
			for (i = 0; i < (tdbp->tdb_key_bits_a + 7)/ 8; i++) {
				akp[i] ^= 0x36;
			}
			MD5Init(&((struct md5_ctx*)(tdbp->tdb_key_a))->ictx);
			
			MD5Update(&((struct md5_ctx*)(tdbp->tdb_key_a))->ictx,
				  akp,
				  (tdbp->tdb_key_bits_a + 7)/ 8);
			for (; i < AHMD596_BLKLEN; i++) {
				MD5Update(&((struct md5_ctx*)(tdbp->tdb_key_a))->ictx,
					  "\0x36", 1);
			}
			
			MD5Init(&((struct md5_ctx*)(tdbp->tdb_key_a))->octx);
			for (i = 0; i < (tdbp->tdb_key_bits_a + 7)/ 8; i++) {
				akp[i] ^= (0x36 ^ 0x5c);
			}
			MD5Update(&((struct md5_ctx*)(tdbp->tdb_key_a))->octx,
				  akp,
				  (tdbp->tdb_key_bits_a + 7)/ 8);
			for (; i < AHMD596_BLKLEN; i++) {
				MD5Update(&((struct md5_ctx*)(tdbp->tdb_key_a))->ictx,
					  "\0x5c", 1);
			}
			
#if 0
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_tdb_init: MD5 ictx=0x%08x %08x %08x %08x"
				    " octx=0x%08x %08x %08x %08x\n",
				    *((__u32*)&((struct md5_ctx*)(tdbp->tdb_key_a))->ictx),
				    *(((__u32*)&((struct md5_ctx*)(tdbp->tdb_key_a))->ictx) + 1),
				    *(((__u32*)&((struct md5_ctx*)(tdbp->tdb_key_a))->ictx) + 2),
				    *(((__u32*)&((struct md5_ctx*)(tdbp->tdb_key_a))->ictx) + 3),
				    *((__u32*)&((struct md5_ctx*)(tdbp->tdb_key_a))->octx),
				    *(((__u32*)&((struct md5_ctx*)(tdbp->tdb_key_a))->octx) + 1),
				    *(((__u32*)&((struct md5_ctx*)(tdbp->tdb_key_a))->octx) + 2),
				    *(((__u32*)&((struct md5_ctx*)(tdbp->tdb_key_a))->octx) + 3));
#endif
			/* paranoid */
			memset(akp, 0, (tdbp->tdb_key_bits_a + 7)/ 8);
			break;
#endif /* CONFIG_IPSEC_AUTH_HMAC_MD5 */
#ifdef CONFIG_IPSEC_AUTH_HMAC_SHA1
		case AH_SHA:
			if(tdbp->tdb_key_bits_a != (AHSHA196_KLEN * 8)) {
				KLIPS_PRINT(debug_pfkey,
					    "klips_debug:pfkey_tdb_init: incorrect authorisation"
					    " key size: %d bits -- must be %d bits\n"/*octets (bytes)\n"*/,
					    tdbp->tdb_key_bits_a, AHSHA196_KLEN * 8);
				SENDERR(EINVAL);
			}
			
#if 0
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_tdb_init: hmac sha1-96 key is 0x%08lx %08lx %08lx %08lx\n",
				    ntohl(*(((__u32 *)ed->ame_key)+0)),
				    ntohl(*(((__u32 *)ed->ame_key)+1)),
				    ntohl(*(((__u32 *)ed->ame_key)+2)),
				    ntohl(*(((__u32 *)ed->ame_key)+3)));
#endif
			tdbp->tdb_auth_bits = AHSHA196_ALEN * 8;
			
			/* save the pointer to the key material */
			akp = tdbp->tdb_key_a;

			if((tdbp->tdb_key_a = (caddr_t)
			    kmalloc((tdbp->tdb_key_a_size = sizeof(struct sha1_ctx)),
				    GFP_ATOMIC)) == NULL) {
				SENDERR(ENOMEM);
			}
			for (i = 0; i < (tdbp->tdb_key_bits_a + 7)/ 8; i++) {
				akp[i] ^= 0x36;
			}
			SHA1Init(&((struct sha1_ctx*)(tdbp->tdb_key_a))->ictx);
			
			SHA1Update(&((struct sha1_ctx*)(tdbp->tdb_key_a))->ictx,
				   akp,
				   (tdbp->tdb_key_bits_a + 7)/ 8);
			for (; i < AHSHA196_BLKLEN; i++) {
				SHA1Update(&((struct sha1_ctx*)(tdbp->tdb_key_a))->ictx,
					   "\0x36", 1);
			}
			
			SHA1Init(&((struct sha1_ctx*)(tdbp->tdb_key_a))->octx);
			for (i = 0; i < (tdbp->tdb_key_bits_a + 7)/ 8; i++) {
				akp[i] ^= (0x36 ^ 0x5c);
			}
			SHA1Update(&((struct sha1_ctx*)(tdbp->tdb_key_a))->octx,
				   akp,
				   (tdbp->tdb_key_bits_a + 7)/ 8);
			for (; i < AHSHA196_BLKLEN; i++) {
				SHA1Update(&((struct sha1_ctx*)(tdbp->tdb_key_a))->ictx,
					   "\0x5c", 1);
			}
			
#if 0
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_tdb_init: SHA1 ictx=0x%08x %08x %08x %08x"
				    " octx=0x%08x %08x %08x %08x\n",
				    *((__u32*)&((struct sha1_ctx*)(tdbp->tdb_key_a))->ictx),
				    *(((__u32*)&((struct sha1_ctx*)(tdbp->tdb_key_a))->ictx) + 1),
				    *(((__u32*)&((struct sha1_ctx*)(tdbp->tdb_key_a))->ictx) + 2),
				    *(((__u32*)&((struct sha1_ctx*)(tdbp->tdb_key_a))->ictx) + 3),
				    *((__u32*)&((struct sha1_ctx*)(tdbp->tdb_key_a))->octx),
				    *(((__u32*)&((struct sha1_ctx*)(tdbp->tdb_key_a))->octx) + 1),
				    *(((__u32*)&((struct sha1_ctx*)(tdbp->tdb_key_a))->octx) + 2),
				    *(((__u32*)&((struct sha1_ctx*)(tdbp->tdb_key_a))->octx) + 3));
#endif
			memset(akp, 0, (tdbp->tdb_key_bits_a + 7)/ 8);
			break;
#endif /* CONFIG_IPSEC_AUTH_HMAC_SHA1 */
		case AH_NONE:
			break;
		default:
					SENDERR(EINVAL);
		}
	}
			break;
#endif /* !CONFIG_IPSEC_ESP */
#ifdef CONFIG_IPSEC_COMP
	case IPPROTO_COMP:
#endif /* CONFIG_IPSEC_COMP */
	default:
		SENDERR(EINVAL);
	}
	
 errlab:
	return(error);
}

int
pfkey_getspi_parse(struct sock *sk, void **extensions, struct tdb *tdbp, struct tdb *tdb2)
{
	int error = 0;
	ipsec_spi_t minspi, maxspi;
	int found_avail = 0;
	struct tdb *tdbq;
	struct timeval time;
	char sa[SATOA_BUF];

	if(!tdbp) {
		SENDERR(EINVAL);
	}

	if((tdbp->tdb_state = ((struct sadb_sa *)extensions[SADB_EXT_SA])->sadb_sa_state) !=
	   SADB_SASTATE_LARVAL) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug: pfkey_getspi_parse: EMT_GETSPI request must be LARVAL.\n");
		SENDERR(EINVAL);
	}

	minspi = ((struct sadb_spirange *)extensions[SADB_EXT_SPIRANGE])->sadb_spirange_min;
	maxspi = ((struct sadb_spirange *)extensions[SADB_EXT_SPIRANGE])->sadb_spirange_max;

	if(maxspi <= 255) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug: pfkey_getspi_parse: EMT_GETSPI maxspi=%d must be > 255.\n", maxspi);
		SENDERR(EEXIST);
	}

	if(maxspi == minspi) {
		tdbp->tdb_spi = maxspi;
		if((tdbq = gettdb(&(tdbp->tdb_said)))) {
			satoa(tdbp->tdb_said, 0, sa, SATOA_BUF);
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug: pfkey_getspi_parse: EMT_GETSPI found an old Tunnel Descriptor Block\n"
				    "klips_debug:                for SA: %s, delete it first.\n", sa);
			SENDERR(EEXIST);
		} else {
			found_avail = 1;
		}
	} else {
		int i = 0;
		while( ( i < (maxspi - minspi)) && !found_avail ) {
			get_random_bytes((void*) &(tdbp->tdb_spi),
					 /* sizeof(tdbp->tdb_spi) */
					 ( maxspi - minspi + 256 ) / 256 );
			tdbp->tdb_spi = htonl(minspi +
					      (tdbp->tdb_spi %
					      (maxspi - minspi + 1)));
			i++;
			tdbq = gettdb(&(tdbp->tdb_said));
			if(!tdbq) {
				found_avail = 1;
			}
		}
	}

	satoa(tdbp->tdb_said, 0, sa, SATOA_BUF);

	if (!found_avail) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug: pfkey_getspi_parse: found an old Tunnel Descriptor Block\n"
			    "klips_debug:                for SA: %s, delete it first.\n", sa);
		SENDERR(EEXIST);
	}

	if(ip_chk_addr((unsigned long)tdbp->tdb_dst.s_addr) == IS_MYADDR) {
		tdbp->tdb_flags |= EMT_INBOUND;
	}

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug: pfkey_getspi_parse: existing Tunnel Descriptor Block not found (this \n"
		    "klips_debug:                is good) for SA: %s, %s-bound, allocating.\n",
		    sa, tdbp->tdb_flags & EMT_INBOUND ? "in" : "out");
	
	/* XXX tdbp->tdb_rcvif = &(enc_softc[em->em_if].enc_if);*/
	tdbp->tdb_rcvif = NULL;
	do_gettimeofday(&time);
	tdbp->tdb_lifetime_addtime_c = time.tv_sec;
	puttdb(tdbp);
	
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_getspi_parse: successful for SA: %s\n", sa);
	
 errlab:
	tdbp->tdb_state = SADB_SASTATE_DEAD;
	return error;
}

int
pfkey_update_parse(struct sock *sk, void **extensions, struct tdb *tdbp, struct tdb *tdb2)
{
	int error = 0;
	struct tdb* tdbq;
	char sa[SATOA_BUF];

#if 0
	if(((struct sadb_sa*)extensions[SADB_EXT_SA])->sadb_sa_state != SADB_SASTATE_MATURE) {
		SENDERR(EINVAL);
	}
#endif

	if(!tdbp) {
		SENDERR(EINVAL);
	}

	satoa(tdbp->tdb_said, 0, sa, SATOA_BUF);

	if((((struct sadb_msg*)extensions[SADB_EXT_RESERVED])->sadb_msg_satype == SADB_SATYPE_AH) &&
	   (((struct sadb_sa*)extensions[SADB_EXT_SA])->sadb_sa_auth == SADB_AALG_NONE)) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_update_parse: no auth algo chosen for SA=%s.\n", sa);
		SENDERR(EINVAL);
	}

	tdbq = gettdb(&(tdbp->tdb_said));
	if (!tdbq) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug: pfkey_update_parse: reserved Tunnel Descriptor Block\n"
			    "klips_debug:            for SA: %s not found.  Call SADB_GETSPI first or call SADB_ADD instead.\n", sa);
		SENDERR(EEXIST);
	}

	if(ip_chk_addr((unsigned long)tdbp->tdb_dst.s_addr) == IS_MYADDR) {
		tdbp->tdb_flags |= EMT_INBOUND;
	}

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug: pfkey_update_parse: existing Tunnel Descriptor Block found (this \n"
		    "klips_debug:                is good) for SA: %s, %s-bound, updating.\n",
		    sa, tdbp->tdb_flags & EMT_INBOUND ? "in" : "out");
	
	/* XXX tdbp->tdb_rcvif = &(enc_softc[em->em_if].enc_if);*/
	tdbp->tdb_rcvif = NULL;
	if ((error = pfkey_tdb_init(tdbp, extensions))) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_update_parse: not successful for SA: %s, deleting.\n", sa);
		ipsec_tdbwipe(tdbp);
		SENDERR(-error);
	}

	if((error = deltdbchain(tdbq))) {
		SENDERR(-error);
	}
	puttdb(tdbp);
	ipsec_tdbwipe(tdbq);
	
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_update_parse: successful for SA: %s\n", sa);
	
#if 0
	if ((error = ipsec_makeroute(&(em->em_eaddr), &(em->em_emask), em->em_ersaid)))
		SENDERR(-error);
#endif

 errlab:
	return error;
}

int
pfkey_add_parse(struct sock *sk, void **extensions, struct tdb *tdbp, struct tdb *tdb2)
{
	int error = 0;
	struct tdb* tdbq;
	char sa[SATOA_BUF];
	struct timeval time;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_add_parse: parsing add message\n");

#if 0
	if(((struct sadb_sa*)extensions[SADB_EXT_SA])->sadb_sa_state != SADB_SASTATE_MATURE) {
		SENDERR(EINVAL);
	}
#endif

	if(!tdbp) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_add_parse: tdbp pointer NULL\n");
		SENDERR(EINVAL);
	}

	satoa(tdbp->tdb_said, 0, sa, SATOA_BUF);

	if((((struct sadb_msg*)extensions[SADB_EXT_RESERVED])->sadb_msg_satype == SADB_SATYPE_AH) &&
	   (((struct sadb_sa*)extensions[SADB_EXT_SA])->sadb_sa_auth == SADB_AALG_NONE)) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_add_parse: no auth algo chosen for SA=%s.\n", sa);
		SENDERR(EINVAL);
	}

	tdbq = gettdb(&(tdbp->tdb_said));
	if (tdbq) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_add_parse: found an old Tunnel Descriptor Block\n"
			    "klips_debug:                for SA: %s, delete it first.\n", sa);
		SENDERR(EEXIST);
	}

	if(ip_chk_addr((unsigned long)tdbp->tdb_dst.s_addr) == IS_MYADDR) {
		tdbp->tdb_flags |= EMT_INBOUND;
	}

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_add_parse: existing Tunnel Descriptor Block not found (this \n"
		    "klips_debug:                is good) for SA: %s, %s-bound, allocating.\n",
		    sa, tdbp->tdb_flags & EMT_INBOUND ? "in" : "out");
	
	/* XXX tdbp->tdb_rcvif = &(enc_softc[em->em_if].enc_if);*/
	tdbp->tdb_rcvif = NULL;
	do_gettimeofday(&time);
	tdbp->tdb_lifetime_addtime_c = time.tv_sec;
	
	if ((error = pfkey_tdb_init(tdbp, extensions))) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_add_parse: not successful for SA: %s, deleting.\n", sa);
		ipsec_tdbwipe(tdbp);
		SENDERR(-error);
	}
	puttdb(tdbp);
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_add_parse: successful for SA: %s\n", sa);
	
#if 0
	if ((error = ipsec_makeroute(&(em->em_eaddr), &(em->em_emask), em->em_ersaid)))
		SENDERR(-error);
#endif
 errlab:
	/* XXX send a message up about success/failure */
	return error;
}

int
pfkey_delete_parse(struct sock *sk, void **extensions, struct tdb *tdbq, struct tdb *tdb2)
{
	struct tdb *tdbp;
	char sa[SATOA_BUF];
	int error = 0;

	satoa(tdbq->tdb_said, 0, sa, SATOA_BUF);

	tdbp = gettdb(&(tdbq->tdb_said));
	if (tdbp == NULL) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_delete_parse: Tunnel Descriptor Block not found for SA:\n"
			    "klips_debug:                %s, could not delete.\n", sa);
		SENDERR(ENXIO);  /* XXX -- wrong error message... */
	} else {
		if((error = deltdbchain(tdbp))) {
			SENDERR(-error);
		}
	}
 errlab:
	ipsec_tdbwipe(tdbq);
	return error;
}

int
pfkey_get_parse(struct sock *sk, void **extensions, struct tdb *tdb1, struct tdb *tdb2)
{
	int error = 0;

	SENDERR(ENOSYS);
 errlab:
	return error;
}

int
pfkey_acquire_parse(struct sock *sk, void **extensions, struct tdb *tdb1, struct tdb *tdb2)
{
	int error = 0;

	SENDERR(ENOSYS);
 errlab:
	return error;
}

int
pfkey_register_parse(struct sock *sk, void **extensions, struct tdb *tdb1, struct tdb *tdb2)
{
	int error = 0;

	SENDERR(ENOSYS);
 errlab:
	return error;
}

int
pfkey_expire_parse(struct sock *sk, void **extensions, struct tdb *tdb1, struct tdb *tdb2)
{
	int error = 0;

	SENDERR(ENOSYS);
 errlab:
	return error;
}

int
pfkey_flush_parse(struct sock *sk, void **extensions, struct tdb *tdb1, struct tdb *tdb2)
{
	int error = 0;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug: pfkey_flush_parse: flushing type %d SAs\n",
		    ((struct sadb_msg*)extensions[SADB_EXT_RESERVED])->sadb_msg_satype);
	if ((error = ipsec_tdbcleanup(((struct sadb_msg*)extensions[SADB_EXT_RESERVED])->sadb_msg_satype)))
		SENDERR(-error);
 errlab:
	return error;
}

int
pfkey_dump_parse(struct sock *sk, void **extensions, struct tdb *tdb1, struct tdb *tdb2)
{
	int error = 0;

	SENDERR(ENOSYS);
 errlab:
	return error;
}

int
pfkey_x_promisc_parse(struct sock *sk, void **extensions, struct tdb *tdb1, struct tdb *tdb2)
{
	int error = 0;

	SENDERR(ENOSYS);
 errlab:
	return error;
}

int
pfkey_x_pchange_parse(struct sock *sk, void **extensions, struct tdb *tdb1, struct tdb *tdb2)
{
	int error = 0;

	SENDERR(ENOSYS);
 errlab:
	return error;
}

int
pfkey_x_grpsa_parse(struct sock *sk, void **extensions, struct tdb *tdb1, struct tdb *tdb2)
{
	struct tdb *tdb1p, *tdb2p, *tdbp;
	char sa1[SATOA_BUF], sa2[SATOA_BUF];
	int error = 0;

	if(!tdb1 || !tdb2) {
		SENDERR(EINVAL);
	}

	satoa(tdb1->tdb_said, 0, sa1, SATOA_BUF);
	satoa(tdb2->tdb_said, 0, sa2, SATOA_BUF);

	if(!(tdb1p = gettdb(&(tdb1->tdb_said)))) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug: pfkey_x_grpsa_parse: reserved Tunnel Descriptor Block\n"
			    "klips_debug:            for SA: %s not found.  Call SADB_ADD/UPDATE first.\n", sa1);
		SENDERR(EEXIST);
	}
	if(!(tdb2p = gettdb(&(tdb2->tdb_said)))) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug: pfkey_x_grpsa_parse: reserved Tunnel Descriptor Block\n"
			    "klips_debug:            for SA: %s not found.  Call SADB_ADD/UPDATE first.\n", sa2);
		SENDERR(EEXIST);
	}

	/* Is tdb1 already linked to tdb2? */
	tdbp = tdb2p;
	while(tdbp) {
		if(tdbp == tdb1p) {
			SENDERR(EEXIST);
		}
		tdbp = tdb2p->tdb_onext;
	}

	/* Is either one already linked? */
	if(tdb1p->tdb_onext) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug: pfkey_x_grpsa_parse: Tunnel Descriptor Block\n"
			    "klips_debug:                for SA: %s is already linked.\n", sa1);
		SENDERR(EEXIST);
	}
	if(tdb2p->tdb_inext) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug: pfkey_x_grpsa_parse: Tunnel Descriptor Block\n"
			    "klips_debug:                for SA: %s is already linked.\n", sa2);
		SENDERR(EEXIST);
	}

	/* link 'em */
	tdb1p->tdb_onext = tdb2p;
	tdb2p->tdb_inext = tdb1p;
 errlab:
	ipsec_tdbwipe(tdb1);
	memset((caddr_t)tdb1, 0, sizeof(*tdb1));
	kfree_s(tdb1, sizeof(*tdb1));
	tdb1 = NULL;

	ipsec_tdbwipe(tdb2);
	memset((caddr_t)tdb2, 0, sizeof(*tdb2));
	kfree_s(tdb2, sizeof(*tdb2));
	tdb1 = NULL;

	return error;
}

int (*msg_parsers[SADB_MAX +1])(struct sock *sk, void *extensions[], struct tdb *tdb1, struct tdb *tdb2)
 =
{
	NULL, /* RESERVED */
	pfkey_getspi_parse,
	pfkey_update_parse,
	pfkey_add_parse,
	pfkey_delete_parse,
	pfkey_get_parse,
	pfkey_acquire_parse,
	pfkey_register_parse,
	pfkey_expire_parse,
	pfkey_flush_parse,
	pfkey_dump_parse,
	pfkey_x_promisc_parse,
	pfkey_x_pchange_parse,
	pfkey_x_grpsa_parse
};

int
pfkey_msg_parse(struct sock *sk, struct sadb_msg *pfkey_msg)
{
	int error = 0;
	int remain;
	struct sadb_ext *pfkey_ext;
	struct tdb *tdb1, *tdb2;
	int extensions_seen = 0;
	void * extensions[SADB_EXT_MAX + 1];

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_msg_parse: parsing message "
		    "ver=%d, type=%d, errno=%d, satype=%d, len=%d, res=%d, seq=%d, pid=%d.\n", 
		    pfkey_msg->sadb_msg_version,
		    pfkey_msg->sadb_msg_type,
		    pfkey_msg->sadb_msg_errno,
		    pfkey_msg->sadb_msg_satype,
		    pfkey_msg->sadb_msg_len,
		    pfkey_msg->sadb_msg_reserved,
		    pfkey_msg->sadb_msg_seq,
		    pfkey_msg->sadb_msg_pid);

	memset((caddr_t)&extensions, 0, sizeof(extensions));
	
	remain = pfkey_msg->sadb_msg_len;
	remain -= sizeof(struct sadb_msg) / IPSEC_PFKEYv2_ALIGN;
	
	pfkey_ext = (struct sadb_ext*)((void*)pfkey_msg +
				       sizeof(struct sadb_msg));
	
#if 0
	if(pfkey_msg->sadb_msg_type == SADB_ADD) {
#endif
		/* check for and use tdbp from previous ACQUIRE or GETSPI call? XXX */
		if((tdb1 = kmalloc(sizeof(*tdb1), GFP_ATOMIC) ) == NULL) {
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_msg_parse: memory allocation error\n");
			SENDERR(ENOMEM);
		}
		memset((caddr_t)tdb1, 0, sizeof(*tdb1));
		if((tdb2 = kmalloc(sizeof(*tdb2), GFP_ATOMIC) ) == NULL) {
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_msg_parse: memory allocation error\n");
			SENDERR(ENOMEM);
		}
		memset((caddr_t)tdb2, 0, sizeof(*tdb2));
#if 0
	}
#endif
	
	extensions[0] = pfkey_msg;

	if(pfkey_msg->sadb_msg_satype > SADB_SATYPE_MAX) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_msg_parse: satype %d > max %d\n", 
			    pfkey_msg->sadb_msg_type, SADB_SATYPE_MAX);
		SENDERR(EINVAL);
	}

	switch(pfkey_msg->sadb_msg_type) {
	case SADB_GETSPI:
	case SADB_UPDATE:
	case SADB_ADD:
	case SADB_DELETE:
	case SADB_X_GRPSA:
		if(!(tdb1->tdb_proto = sadb_satype2proto[pfkey_msg->sadb_msg_satype])) {
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_msg_parse: satype %d translates to zero\n", 
				    pfkey_msg->sadb_msg_satype);
			SENDERR(EINVAL);
		}
		break;
	default:
	}

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_msg_parse: remain=%d, ext_type=%d, ext_len=%d.\n", 
		    remain, pfkey_ext->sadb_ext_type, pfkey_ext->sadb_ext_len);

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_msg_parse: extensions "
		    "permitted=%08x, required=%08x.\n",
		    exts_permitted_in[pfkey_msg->sadb_msg_type],
		    exts_required_in[pfkey_msg->sadb_msg_type]);

	extensions_seen =  1 ;

	while(remain && (remain >= sizeof(struct sadb_ext) / IPSEC_PFKEYv2_ALIGN)) {
		/* Is there enough message left to support another extension header? */
		if(remain < pfkey_ext->sadb_ext_len) {
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_msg_parse: remain %d less than ext len %d.\n", 
				    remain, pfkey_ext->sadb_ext_len);
			SENDERR(EINVAL);
		}

		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_msg_parse: parsing ext type=%d remain=%d.\n",
			    pfkey_ext->sadb_ext_type,
			    remain);

		/* Is the extension header type valid? */
		if((pfkey_ext->sadb_ext_type > SADB_EXT_MAX) || (!pfkey_ext->sadb_ext_type)) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_msg_parse: ext type %d invalid, SADB_EXT_MAX=%d.\n", 
			    pfkey_ext->sadb_ext_type, SADB_EXT_MAX);
			SENDERR(EINVAL);
		}

		/* Have we already seen this type of extension? */
		if(extensions[pfkey_ext->sadb_ext_type]) {
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_msg_parse: ext type %d already seen.\n", 
				    pfkey_ext->sadb_ext_type);
			SENDERR(EINVAL);
		}

		/* Is this type of extension permitted for this type of message? */
		if(!(exts_permitted_in[pfkey_msg->sadb_msg_type] & 1<<pfkey_ext->sadb_ext_type)) {
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_msg_parse: ext type %d not permitted, exts_perm_in=%08x, 1<<type=%08x\n", 
				    pfkey_ext->sadb_ext_type, 
				    exts_permitted_in[pfkey_msg->sadb_msg_type],
				    1<<pfkey_ext->sadb_ext_type);
			SENDERR(EINVAL);
		}

		/* Parse the extension */
		if((error = ext_parsers[pfkey_ext->sadb_ext_type](sk, pfkey_msg, pfkey_ext, tdb1, tdb2))) {
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_msg_parse: ext_parse returns %d.\n", 
				    error);
			SENDERR(-error);
		}

		/* Mark that we have seen this extension and remember the header location */
		extensions_seen |= ( 1 << pfkey_ext->sadb_ext_type );
		extensions[pfkey_ext->sadb_ext_type] = pfkey_ext;

		/* Calculate how much message remains */
		remain -= pfkey_ext->sadb_ext_len;

		if(!remain) {
			break;
		}
		/* Find the next extension header */
		pfkey_ext = (struct sadb_ext*)((void*)pfkey_ext +
			pfkey_ext->sadb_ext_len * IPSEC_PFKEYv2_ALIGN);
	}

	if(remain) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_msg_parse: unexpected remainder of %d.\n", 
			    remain);
		/* why is there still something remaining? */
		SENDERR(EINVAL);
	}

	/* check required extensions */
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_msg_parse: extensions "
		    "permitted=%08x, seen=%08x, required=%08x.\n",
		    exts_permitted_in[pfkey_msg->sadb_msg_type],
		    extensions_seen,
		    exts_required_in[pfkey_msg->sadb_msg_type]);

	if((extensions_seen & exts_required_in[pfkey_msg->sadb_msg_type]) !=
	   exts_required_in[pfkey_msg->sadb_msg_type]) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_msg_parse: required extensions missing.\n");
		SENDERR(EINVAL);
	}
	
	/* Parse the message */
	if((error = msg_parsers[pfkey_msg->sadb_msg_type](sk, extensions, tdb1, tdb2))) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_msg_parse: message parsing failed with error %d.\n",
			    error); 
		SENDERR(-error);
	}

errlab:
	/* XXX free tdb1, tdb2 */
	return error;
}

/*
 * $Log: pfkey_v2_parser.c,v $
 * Revision 1.10  1999/09/21 15:24:45  rgb
 * Rework spirange code to save entropy and prevent endless loops.
 *
 * Revision 1.9  1999/09/16 12:10:21  rgb
 * Minor fixes to random spi selection for correctness and entropy conservation.
 *
 * Revision 1.8  1999/05/25 22:54:46  rgb
 * Fix comparison that should be an assignment in an if.
 *
 * Revision 1.7  1999/05/09 03:25:37  rgb
 * Fix bug introduced by 2.2 quick-and-dirty patch.
 *
 * Revision 1.6  1999/05/08 21:32:30  rgb
 * Fix error return reporting.
 *
 * Revision 1.5  1999/05/05 22:02:33  rgb
 * Add a quick and dirty port to 2.2 kernels by Marc Boucher <marc@mbsi.ca>.
 *
 * Revision 1.4  1999/04/29 15:22:40  rgb
 * Standardise an error return method.
 * Add debugging instrumentation.
 * Add check for existence of macros min/max.
 * Add extensions permitted/required in/out filters.
 * Add satype-to-protocol table.
 * Add a second tdb pointer to each parser to accomodate GRPSA.
 * Move AH & no_algo_set to GETSPI, UPDATE and ADD.
 * Add OOO window check.
 * Add support for IPPROTO_IPIP and hooks for IPPROTO_COMP.
 * Add timestamp to lifetime parse.
 * Fix address structure length checking bug.
 * Fix address structure allocation bug (forgot to kmalloc!).
 * Add checks for extension lengths.
 * Add checks for extension reserved illegal values.
 * Add check for spirange legal values.
 * Add an extension type for parsing a second satype, SA and
 * DST_ADDRESS.
 * Make changes to tdb_init() template to get pfkey_tdb_init(),
 * eliminating any mention of xformsw.
 * Implement getspi, update and grpsa (not tested).
 * Add stubs for as yet unimplemented message types.
 * Add table of message parsers to substitute for msg_parse switch.
 *
 * Revision 1.3  1999/04/15 17:58:07  rgb
 * Add RCSID labels.
 *
 * Revision 1.2  1999/04/15 15:37:26  rgb
 * Forward check changes from POST1_00 branch.
 *
 * Revision 1.1.2.1  1999/03/26 20:58:56  rgb
 * Add pfkeyv2 support to KLIPS.
 *
 */
